DEBUG = False
DEBUG_OVERFIT = False
DEBUG_SPARSE = False
DEBUG_ORACLE_LENGTH = False
DEBUG_DIFF_LR = False
DEBUG_NON_PREFIX_FOR_LEN = False
DEBUG_PENULTIMATE_LAYER = False
DEBUG_STOPPING = False
DEBUG_OWT = False
DEBUG_CD_NAN = False
DEBUG_OVERFIT_TRAIN = False
DEBUG_XLNET_DATA = False
DEBUG_XLNET_AR = False
DEBUG_PROFILE = False
DEBUG_VIZ_LOGITS = False
DEBUG_PRINT_PREDS = False
